import asyncio
import os
from pathlib import Path

import typer
from rich.console import Console

from . import helpers
from .client import Client
from .constants import Messages
from .models import Quality
from .utils import appdirs, tools

app = typer.Typer(
    rich_markup_mode="markdown",
    epilog="Made with :heart: by [ivansaul](https://github.com/ivansaul)",
)
console = Console()


@app.command()
def login() -> None:
    """
    Authenticates a user with the given email and password.
    """
    while True:
        email = typer.prompt("What's your email?")
        confirm_email = typer.prompt("Confirm your email?")
        if email == confirm_email:
            break
        console.print(Messages.VALUE_DOES_NOT_MATCH)

    while True:
        password = typer.prompt("What's your password?", hide_input=False)
        confirm_password = typer.prompt("Confirm your password?", hide_input=False)
        if password == confirm_password:
            break
        console.print(Messages.VALUE_DOES_NOT_MATCH)

    console.print("⠹ Logging in...", style="bold green")

    try:
        asyncio.run(_login(email, password))
        console.print("✓ Logged in successfully.", style="bold green")
    except Exception as e:
        error_message = str(e)
        console.print(error_message, style="bold red")
        raise typer.Exit()


@app.command()
def logout() -> None:
    """
    Logs out the user.
    """
    appdirs.delete_app_config_file()
    console.print("✓ Logged out successfully.", style="bold green")


@app.command()
def refresh() -> None:
    """
    Refreshes the cookies of the user.
    """
    config_path: Path = Path(appdirs.get_config_file_path())
    if not config_path.is_file():
        console.print("✗ Not logged in. Please login first.", style="bold red")
        raise typer.Exit()

    try:
        credentials = helpers.read_json(config_path.as_posix())
        email = credentials["email"]
        password = credentials["password"]
    except Exception:
        console.print("✗ Not logged in. Please login first.", style="bold red")
        raise typer.Exit()

    console.print("⠹ Refreshing cookies...", style="bold green")

    try:
        asyncio.run(_login(email, password))
        console.print("✓ Cookies refreshed successfully.", style="bold green")
    except Exception as e:
        error_message = str(e)
        console.print(error_message, style="bold red")
        raise typer.Exit()


@app.command()
def download() -> None:
    """
    Downloads the course with the given URL.
    """
    url = typer.prompt("What's the URL of the course?")
    asyncio.run(_download(url))


async def _login(email: str, password: str) -> None:
    async with Client(headless=False) as client:
        await client.login(email, password)


async def _download(url: str) -> None:
    async with Client(headless=False) as client:
        # TODO: implement human verification
        await client.verify_captcha(url)

        info = await client.fetch_course_info(url)
        console.print(info.dict())

        course = await client.fetch_course(url)

        save_dir = course.title
        os.makedirs(save_dir, exist_ok=True)
        for idx, section in enumerate(course.sections, start=1):
            section_dir = os.path.join(save_dir, f"{idx:02}. {section.title}")
            os.makedirs(section_dir, exist_ok=True)

            # dl section videos
            for idx, video in enumerate(section.videos, start=1):
                await tools.download_m3u8_video(
                    url=video.m3u8_url,
                    save_dir=section_dir,
                    quality=Quality.P224,
                    subtitle_lang="en",
                    output_name=f"{idx:02}. {video.title}",
                )

            # dl section assets
            for asset in section.assets:
                output = asset.name + os.path.splitext(asset.url)[1]
                await tools.download_file(asset.url, section_dir, output)

        # dl course assets
        for asset in course.assets:
            output = asset.name + os.path.splitext(asset.url)[1]
            await tools.download_file(asset.url, save_dir, output)

        # dl course data json
        helpers.write_json(course.dict(), os.path.join(save_dir, "course.json"))



class Constants:
    APP_NAME = "Domestica"
    CONFIG_FILE = "config.json"
    COOKIES_FILE = "cookies.json"


class Messages:
    VALUE_DOES_NOT_MATCH = "[bold red]Error![/bold red] [magenta]The two entered values do not match.[/magenta]"



import asyncio

from playwright.async_api import BrowserContext, Page, async_playwright

from .models import Course, CourseInfo
from .utils import appdirs, auth, collectors


class Client:
    def __init__(self, headless: bool = False):
        self.headless = headless

    async def __aenter__(self):
        self._playwright = await async_playwright().start()

        self._browser = await self._playwright.chromium.launch(headless=self.headless)
        self._context = await self._browser.new_context(
            java_script_enabled=True,
            is_mobile=True,
        )

        try:
            await self.load_state()
        except Exception:
            raise Exception("Could not load state")

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._context.close()
        await self._browser.close()
        await self._playwright.stop()

    @property
    def config_path(self) -> str:
        return appdirs.get_config_file_path()

    @property
    async def page(self) -> Page:
        return await self._context.new_page()

    @property
    def context(self) -> BrowserContext:
        return self._context

    async def verify_captcha(self, url: str) -> None:
        # TODO: improve this - human verification
        page = await self.page
        await page.goto(url)
        await asyncio.sleep(20)
        await page.close()

    async def login(self, email: str, password: str):
        await auth.login(email, password, self.context)

    async def load_state(self) -> None:
        await auth.load_state(self.context)

    async def fetch_course_info(self, url: str) -> CourseInfo:
        return await collectors.fetch_course_info(url, self.context)

    async def fetch_course(self, url: str) -> Course:
        return await collectors.fetch_course(url, self.context)



from enum import Enum
from typing import Optional

from pydantic import BaseModel


class Quality(Enum):
    """Video quality"""

    P2160 = "2160"
    P1080 = "1080"
    P720 = "720"
    P540 = "540"
    P360 = "360"
    P224 = "224"


class Video(BaseModel):
    """video model"""

    id: Optional[str] = None
    title: str
    m3u8_url: Optional[str]


class Media(BaseModel):
    """media model"""

    name: str
    url: str


class Section(BaseModel):
    """section model"""

    id: Optional[str] = None
    title: str
    videos: list[Video]
    assets: list[Media] = []


class CourseInfo(BaseModel):
    """course info model"""

    product_id: int
    product_name: str
    teacher_name: str
    category_id: int
    category_name: str
    course_level: str
    course_number_of_lessons: int
    course_total_duration_sec: int
    total_units: int
    subtitles_language: list[str]
    audio_language: list[str]
    original_language: str

    @property
    def slug(self) -> str:
        product_name = self.product_name.strip().lower().replace(" ", "-")
        return f"{self.product_id}-{product_name}"

    @property
    def url(self) -> str:
        return f"https://www.domestika.org/courses/{self.slug}/course"


class Course(BaseModel):
    """course model"""

    id: str
    title: str
    url: str
    sections: list[Section]
    assets: list[Media] = []
    info: Optional[CourseInfo] = None



from pathlib import Path

import typer

from ..constants import Constants


def get_app_dir() -> str:
    """
    Returns the path to the directory where the given app is installed.

    Returns:
        str: The path to the directory where the app is installed.
    """
    app_dir = typer.get_app_dir(Constants.APP_NAME)
    return app_dir


def get_config_file_path() -> str:
    """
    Returns the path to the configuration file for the given app.

    Returns:
        str: The path to the configuration file.
    """
    app_dir = get_app_dir()
    config_path: Path = Path(app_dir) / Constants.CONFIG_FILE
    return config_path.as_posix()


def delete_app_config_file() -> None:
    """
    Deletes the configuration file for the given app.

    Args:
        app_name (str): The name of the app.
    """
    config_path: Path = Path(get_config_file_path())
    if config_path.is_file():
        config_path.unlink()



import asyncio
import os
import shutil
import sys
import tarfile
import zipfile

import aiohttp

from ..models import Quality
from . import appdirs


async def download_m3u8_video(
    url: str,
    save_dir: str,
    quality: Quality = Quality.P1080,
    subtitle_lang: str = "en",
    output_name: str = "output",
) -> None:
    """
    Downloads a video using N_m3u8DL-RE with specific options.

    Parameters:
    - url (str): The URL of the video to be downloaded.
    - save_dir (str): The directory where the downloaded video will be saved.
    - quality (Quality): The desired video quality, e.g., Quality.P1080. Default is Quality.P1080.
    - subtitle_lang (str): The language of the subtitles desired, e.g., "en".
    - output_name (str): The name of the output file without the extension.
    """
    os.makedirs(save_dir, exist_ok=True)

    print(f"[DOWNLOADING] {output_name}", flush=True, end="")

    download_command = [
        "N_m3u8DL-RE",
        "--select-video",
        f"res='{quality.value}*':codec=hvc1:for=best",
        "--sub-format",
        "SRT",
        "--select-subtitle",
        f'lang="{subtitle_lang}":for=all',
        "--auto-subtitle-fix",
        url,
        "--save-dir",
        save_dir,
        "--save-name",
        output_name,
        "--concurrent-download",
    ]

    process = await asyncio.create_subprocess_exec(
        *download_command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    await process.wait()

    if process.returncode == 0:
        print(f"\r[COMPLETED] {output_name}")
    else:
        print(f"\r[ERROR] {output_name}")


async def download_file(url: str, save_dir: str, output: str) -> None:
    """
    Downloads a file from a URL and saves it to the specified path.

    Args:
        url (str): The URL of the file to be downloaded.
        save_dir (str): The path where the file should be saved.
        output (str): The name of the output file.
    """
    os.makedirs(save_dir, exist_ok=True)
    path = os.path.join(save_dir, output)
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            if response.status == 200:
                with open(path, "wb") as file:
                    while True:
                        chunk = await response.content.read(1024)
                        if not chunk:
                            break
                        file.write(chunk)
            else:
                raise Exception(f"Failed to download file: {url}")


async def upload_from_path(path: str) -> str:
    """
    Uploads a file to the catbox.moe service.

    Args:
        path (str): The path to the file to be uploaded.

    Returns:
        str: The response from the catbox.moe service.
    """
    if not os.path.exists(path):
        raise FileNotFoundError(f"{path} does not exist.")

    ENDPOINT = "https://catbox.moe/user/api.php"
    user_hash = "4b6585b7d61940d8b98d8e0a4"
    payload = aiohttp.FormData()
    payload.add_field("reqtype", "fileupload")
    payload.add_field("userhash", user_hash)

    with open(path, "rb") as file:
        payload.add_field(
            "fileToUpload",
            file,
            filename=os.path.basename(path),
        )
        async with aiohttp.ClientSession() as session:
            async with session.post(ENDPOINT, data=payload) as response:
                if response.status == 200:
                    return await response.text()
                else:
                    raise Exception(f"Failed to upload file: {path}")


async def upload_from_bytes(file_data: bytes, filename: str) -> str:
    """
    Uploads a file to the catbox.moe service from bytes.

    Args:
        file_data (bytes): The file data in bytes to be uploaded.
        filename (str): The name of the file to be uploaded.

    Returns:
        str: The response from the catbox.moe service.
    """
    ENDPOINT = "https://catbox.moe/user/api.php"
    user_hash = "4b6585b7d61940d8b98d8e0a4"

    payload = aiohttp.FormData()
    payload.add_field("reqtype", "fileupload")
    payload.add_field("userhash", user_hash)
    payload.add_field(
        "fileToUpload",
        file_data,
        filename=filename,
        content_type="application/octet-stream",
    )

    async with aiohttp.ClientSession() as session:
        async with session.post(ENDPOINT, data=payload) as response:
            if response.status == 200:
                return await response.text()
            else:
                raise Exception(f"Failed to upload file: {filename}")


async def download_dependencies() -> None:
    """
    Downloads the required dependencies for the application.
    """
    app_dir = appdirs.get_app_dir()
    bin_dir = os.path.join(app_dir, "bin")
    dependencies_dir = os.path.join(app_dir, "dependencies")

    os.makedirs(app_dir, exist_ok=True)
    os.makedirs(bin_dir, exist_ok=True)
    os.makedirs(dependencies_dir, exist_ok=True)

    version = "v0.2.1-beta"
    name = "N_m3u8DL-RE_Beta"
    build = "20240828"
    format = "tar.gz"
    platform = "osx-arm64"

    m3u8dl_bin_path = os.path.join(bin_dir, "N_m3u8DL-RE")
    # TODO: add support for arm architecture
    if sys.platform.startswith("darwin"):
        pass

    elif sys.platform.startswith("linux"):
        platform = "linux-x64"

    elif sys.platform.startswith("win"):
        platform = "win-x64"
        format = "zip"
        m3u8dl_bin_path = os.path.join(bin_dir, "N_m3u8DL-RE.exe")

    else:
        raise RuntimeError("Unsupported platform")

    m3u8dl_url = f"https://github.com/nilaoda/N_m3u8DL-RE/releases/download/{version}/{name}_{platform}_{build}.{format}"

    # Download m3u8dl binary
    if not os.path.exists(m3u8dl_bin_path):
        zip_name = f"{name}_{version}.{format}"
        await download_file(m3u8dl_url, dependencies_dir, zip_name)

        zip_path = os.path.join(dependencies_dir, zip_name)

        if format == "zip":
            with zipfile.ZipFile(zip_path, "r") as zip:
                zip.extractall(dependencies_dir)

        if format == "tar.gz":
            with tarfile.open(zip_path, "r") as tar:
                tar.extractall(dependencies_dir)

        os.remove(zip_path)

        # Move binary to bin directory
        for dir, subdirs, files in os.walk(dependencies_dir):
            for file in files:
                if file in ["N_m3u8DL-RE", "N_m3u8DL-RE.exe"]:
                    bin_src = os.path.join(dir, file)
                    shutil.move(bin_src, bin_dir)

        # Set executable permission
        if not os.access(m3u8dl_bin_path, os.X_OK):
            os.chmod(m3u8dl_bin_path, 0o744)

    # Add bin directory to PATH
    if "PATH" not in os.environ:
        os.environ["PATH"] = bin_dir
    elif bin_dir not in os.environ["PATH"]:
        os.environ["PATH"] = bin_dir + os.pathsep + os.environ["PATH"]



import json
import re
from pathlib import Path
from typing import Any


def read_json(path: str) -> dict[str, Any]:
    """
    Reads a JSON file and returns a dictionary.

    Args:
        path (str): The path to the JSON file.

    Returns:
        dict[str, Any]: The dictionary containing the JSON data.
    """
    with open(path, "r", encoding="utf-8") as file:
        return json.load(file)


def write_json(data: Any, path: str) -> None:
    """
    Writes a dictionary to a JSON file.

    Args:
        data (dict[str, Any]): The dictionary to be written to the JSON file.
        path (str): The path to the JSON file.
    """

    Path(path).parent.mkdir(parents=True, exist_ok=True)
    with open(path, "w", encoding="utf-8") as file:
        json.dump(data, file, indent=4, ensure_ascii=False)


def clean_string(text: str) -> str:
    """
    Cleans the input string by removing special characters and
    leading/trailing white spaces.

    Args:
        text (str): The input string to be cleaned.

    Returns:
        str: The cleaned string, with special characters removed and
        leading/trailing spaces stripped.
    """
    pattern = r"[ºª]|[^\w\s]"
    return re.sub(pattern, "", text).strip()



from playwright.async_api import BrowserContext

from .. import helpers
from .appdirs import get_config_file_path


async def load_state(context: BrowserContext) -> bool:
    try:
        config_path = get_config_file_path()
        credentials = helpers.read_json(config_path)
        cookies = credentials["cookies"]
        await context.add_cookies(cookies)  # type: ignore

        page = await context.new_page()
        await page.goto("https://www.domestika.org")
        await page.wait_for_selector("span[class='avatar avatar--s']")
        return True
    except Exception:
        return False
    finally:
        await page.close()


async def login(email: str, password: str, context: BrowserContext) -> None:
    page = await context.new_page()
    await page.goto("https://www.domestika.org/auth/login")
    await page.fill("#user_email", email)
    await page.fill("#user_password", password)
    await page.click(".simple-credentials .t-login-button")

    try:
        await page.wait_for_selector("span[class='avatar avatar--s']")
    except Exception:
        raise Exception("Could not login")

    cookies = await context.cookies()
    credentials = {
        "email": email,
        "password": password,
        "cookies": cookies,
    }
    config_path = get_config_file_path()
    helpers.write_json(credentials, config_path)



import asyncio
import json
import re
from typing import Optional

from playwright.async_api import BrowserContext, Page

from .. import helpers
from ..models import Course, CourseInfo, Media, Section, Video
from ..utils import tools


def get_course_id(url: str) -> str:
    pattern = r"/(\d+)-"
    match = re.search(pattern, url)
    if not match:
        raise Exception("Could not get course id")
    return match.group(1)


def _parse_course_info(content: str) -> CourseInfo:
    pattern = r"window\.Domestika\.courses_controller\.course\((.*?)\);"
    match = re.search(pattern, content)
    if not match:
        raise Exception("Could not get course info")
    try:
        json_str = match.group(1)
        json_data = json.loads(json_str)
        json_info = json_data["amplitude_event"]["event_properties"][
            "product_attributes"
        ][0]
        return CourseInfo(**json_info)
    except Exception as e:
        raise Exception("Could not parse course info") from e


async def fetch_course_info(url: str, context: BrowserContext) -> CourseInfo:
    page = await context.new_page()
    await page.goto(url)
    return _parse_course_info(await page.content())


async def get_course_title(page: Page) -> str:
    try:
        title = await page.locator("h1.course-header-new__title a").text_content()
    except TimeoutError:
        raise Exception("Could not get course title")

    if title is None:
        raise Exception("Could not get course title")
    return helpers.clean_string(title)


async def page_to_img(page: Page, path: Optional[str] = None) -> bytes:
    await page.emulate_media(media="screen")
    return await page.screenshot(path=path, full_page=True)


async def page_to_pdf(page: Page, path: Optional[str] = None) -> bytes:
    await page.emulate_media(media="screen")
    return await page.pdf(path=path)


async def _exit_multi_courses(page: Page) -> None:
    multi_courses_locator = page.locator(".a-tag.bg-color-guijarro")
    if await multi_courses_locator.count() > 0:
        await page.close()
        raise Exception("This kind of course is not supported")


async def fetch_course(url: str, context: BrowserContext) -> Course:
    page = await context.new_page()
    await page.goto(url)

    await _exit_multi_courses(page)

    title = await get_course_title(page)

    try:
        tasks = []
        sections: list[Section] = []
        sections_locator = page.locator("ul.units-list h4.unit-item__title a")
        for i in range(await sections_locator.count()):
            section_url = await sections_locator.nth(i).get_attribute("href")
            if section_url is None:
                raise Exception("Could not get section url")
            tasks.append(fetch_section(section_url, context))
        sections = await asyncio.gather(*tasks)
    except TimeoutError:
        await page.close()
        raise Exception("Could not get sections")
    except Exception as e:
        await page.close()
        raise e

    # Capture content as PDF
    pdf_content_url = await tools.upload_from_bytes(
        await page_to_pdf(page),
        f"{title}_content.pdf",
    )

    info = _parse_course_info(await page.content())

    await page.close()

    return Course(
        id=get_course_id(url),
        title=title,
        url=url,
        sections=sections,
        assets=[
            Media(name="content", url=pdf_content_url),
        ],
        info=info,
    )


async def fetch_section(url: str, context: BrowserContext) -> Section:
    page = await context.new_page()
    await page.goto(url)

    # TODO: implement this
    if url.endswith("/final_project"):
        await page.close()
        return Section(
            title="Final Project",
            videos=[],
            assets=[],
        )
    try:
        section_title = await page.locator("header.paper__header h2").text_content()
    except TimeoutError:
        await page.close()
        raise Exception("Could not get section title")

    if section_title is None:
        await page.close()
        raise Exception("Could not get section title")

    section_title = helpers.clean_string(section_title)

    pattern = r"window\.__INITIAL_PROPS__ = JSON\.parse\('(.*?)'\)"
    match = re.search(pattern, await page.content())
    if not match:
        await page.close()
        raise Exception("Could not get section content")

    json_str = match.group(1).replace("\\", "")
    json_data = json.loads(json_str)
    videos: list[Video] = []
    try:
        for video in json_data["videos"]:
            media = Video(
                title=helpers.clean_string(video["video"]["title"]),
                m3u8_url=video["video"]["playbackURL"],
            )
            videos.append(media)
    except KeyError:
        await page.close()
        raise Exception("Could not get section videos")

    # Capture content as PDF
    pdf_content_url = await tools.upload_from_bytes(
        await page_to_pdf(page),
        f"{section_title}_content.pdf",
    )

    await page.close()

    return Section(
        title=section_title,
        videos=videos,
        assets=[
            Media(name="content", url=pdf_content_url),
        ],
    )
